<?php

namespace Drupal\webform_googlesheets\Plugin\WebformHandler;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\file\Entity\File;
use Drupal\webform\Plugin\WebformElement\WebformManagedFileBase;
use Drupal\webform\Element\WebformMessage;
use Drupal\webform\Plugin\WebformElement\BooleanBase;
use Drupal\webform\Plugin\WebformElement\NumericBase;
use Drupal\webform\Plugin\WebformElement\WebformCompositeBase;
use Drupal\webform\Plugin\WebformElementInterface;
use Drupal\webform\Plugin\WebformHandlerBase;
use Drupal\webform\WebformSubmissionInterface;
use Exception;
use Google\Service\Sheets;
use Google\Service\Sheets\AppendCellsRequest;
use Google\Service\Sheets\BatchUpdateSpreadsheetRequest;
use Google\Service\Sheets\CellData;
use Google\Service\Sheets\ExtendedValue;
use Google\Service\Sheets\Request;
use Google\Service\Sheets\RowData;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form submission to Google Sheets handler.
 *
 * @WebformHandler(
 *   id = "googlesheets",
 *   label = @Translation("Google Sheets"),
 *   category = @Translation("Google Sheets"),
 *   description = @Translation("Adds a form submission to a Google Sheets."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_UNLIMITED,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_PROCESSED,
 *   submission = \Drupal\webform\Plugin\WebformHandlerInterface::SUBMISSION_OPTIONAL,
 *   tokens = TRUE,
 * )
 */
class WebformGoogleSheetsHandler extends WebformHandlerBase {

  /**
   * A webform element plugin manager.
   *
   * @var \Drupal\webform\Plugin\WebformElementManagerInterface
   */
  protected $elementManager;

  /**
   * Google API Client service.
   *
   * @var \Drupal\google_api_client\Service\GoogleApiClientService
   */
  protected $googleService;

  /**
   * List of unsupported webform submission properties.
   *
   * The below properties will not being included in a remote post.
   *
   * @var array
   */
  protected $unsupportedProperties = [
    'metatag',
  ];

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->elementManager = $container->get('plugin.manager.webform.element');
    $instance->googleService = $container->get('google_api_client.client');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary() {
    $configuration = $this->getConfiguration();

    return [
      '#settings' => $configuration['settings'],
    ] + parent::getSummary();
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $field_names = array_keys(
      \Drupal::service('entity_field.manager')
        ->getBaseFieldDefinitions('webform_submission')
    );

    return [
      'spreadsheet_url' => '',
      'spreadsheet_id' => '',
      'spreadsheet_sheet_id' => '',
      'google_api_client' => 0,
      'included_data' => array_combine($field_names, $field_names),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    parent::buildConfigurationForm($form, $form_state);

    // Google Sheets configuration.
    $form['googlesheets'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Google Sheets settings'),
    ];
    $form['googlesheets']['spreadsheet_url'] = [
      '#type' => 'url',
      '#title' => $this->t('Sheet URL'),
      '#required' => TRUE,
      '#default_value' => $this->configuration['spreadsheet_url'] ?? '',
      '#empty_value' => '',
    ];

    $form['googlesheets']['google_api_client'] = [
      '#type' => 'select',
      '#title' => $this->t('Google API Client'),
      '#required' => TRUE,
      '#options' => [],
      '#default_value' => $this->configuration['google_api_client'] ?? 0,
      '#empty_value' => '',
    ];
    $google_api_clients = $this->getEntityStorage('google_api_client')->loadByProperties([]);
    foreach ($google_api_clients as $google_api_client) {
      $form['googlesheets']['google_api_client']['#options'][$google_api_client->id()] = $google_api_client->label();
    }


    /** @var \Drupal\google_api_client\Entity\GoogleApiClient $google_api_client */
    $google_api_client = $this->getEntityStorage('google_api_client')->load($this->configuration['google_api_client'] ?? 0);
    if ($google_api_client) {
      $items = [];

      // Check if services are right.
      $correct_service = FALSE;
      $services = $google_api_client->getServices();
      foreach ($services as $service) {
        if ($service == 'sheets') {
          $correct_service = TRUE;
          break;
        }
      }
      if (!$correct_service) {
        $items[] = new TranslatableMarkup('Select %service as a service.', ['%service' => 'Google Sheets API']);
      }

      // Check if scopes are right.
      $correct_scope = FALSE;
      $scopes = $google_api_client->getScopes();
      foreach ($scopes as $scope) {
        if ($scope == 'https://www.googleapis.com/auth/spreadsheets') {
          $correct_scope = TRUE;
          break;
        }
      }
      if (!$correct_scope) {
        $items[] = new TranslatableMarkup('Select %scope as a scope.', ['%scope' => 'https://www.googleapis.com/auth/spreadsheets']);
      }

      // Check if authenticated.
      if (!$google_api_client->getAuthenticated()) {
        $link = Link::fromTextAndUrl(
          new TranslatableMarkup('here'),
          Url::fromRoute('google_api_client.callback', [], [
            'attributes' => ['target' => '_blank'],
            'query' => ['id' => $google_api_client->id()]
          ])
        );
        $items[] = new TranslatableMarkup('Perform authentication @link', ['@link' => $link->toString()]);
      }

      if ($items) {
        $form['googlesheets']['google_api_client_issues'] = [
          '#theme' => 'item_list',
          '#list_type' => 'ul',
          '#title' => $items ? new TranslatableMarkup('%google_api_client client has following issues:', ['%google_api_client' => $google_api_client->label()]) : new TranslatableMarkup('%google_api_client client is correctly configured', ['%google_api_client' => $google_api_client->label()]),
          '#items' => $items,
        ];
      }
    }

    // Submission data.
    $form['submission_data'] = [
      '#type' => 'details',
      '#title' => $this->t('Submission data'),
    ];
    // Display warning about file uploads.
    if ($this->getWebform()->hasManagedFile()) {
      $form['submission_data']['managed_file_message'] = [
        '#type' => 'webform_message',
        '#message_message' => $this->t('Upload files will include the file\'s id, name, uri, and data (<a href=":href">Base64</a> encode).', [':href' => 'https://en.wikipedia.org/wiki/Base64']),
        '#message_type' => 'warning',
        '#message_close' => TRUE,
        '#message_id' => 'webform_node.references',
        '#message_storage' => WebformMessage::STORAGE_SESSION,
      ];
    }
    $form['submission_data']['included_data'] = [
      '#type' => 'webform_excluded_columns',
      '#title' => $this->t('Submitted data'),
      '#title_display' => 'invisible',
      '#webform_id' => $this->getWebform()->id(),
      '#required' => TRUE,
      '#default_value' => $this->configuration['included_data'] ?? [],
    ];

    return $this->setSettingsParents($form);
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::validateConfigurationForm($form, $form_state);

    if (
      !preg_match('#https?://docs.google.com/spreadsheets/d/(.+)/edit(\#gid=(\d+))?#', $form_state->getValue('spreadsheet_url'), $matches)
    ) {
      $form_state->setErrorByName('googlesheets][spreadsheet_url', $this->t('Please provide a valid Google Sheet URL.'));
    }
    else {
      $form_state->setValue('spreadsheet_id', $matches[1]);
      $form_state->setValue('spreadsheet_sheet_id', !empty($matches[3]) ? $matches[3] : 0);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    $this->applyFormStateToConfiguration($form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(WebformSubmissionInterface $webform_submission, $update = TRUE) {
    // If update operation, do nothing.
    if ($update) {
      return;
    }

    // Prepare request.
    $data = $this->getData($webform_submission);
    $rows = $this->prepareRows($data);
    $request = $this->prepareRequest($this->configuration['spreadsheet_sheet_id'], $rows);

    // Execute request.
    $error = FALSE;
    try {
      $this
        ->makeRequest($this->configuration['spreadsheet_id'], $request)
        ->valid();
    }
    catch (\Exception $e) {
      $error = $e->getMessage();
    }

    // Log detailed message to the 'webform_submission' log.
    $context = [
      'link' => ($webform_submission->id()) ? $webform_submission->toLink($this->t('View'))->toString() : NULL,
      'webform_submission' => $webform_submission,
      'handler_id' => $this->getHandlerId(),
      'operation' => 'google sheets append',
      '%error' => $error,
    ];
    if (!$error) {
      $this->getLogger('webform_submission')->notice("Webform submission had been successfully appended to Google Sheets.", $context);
    }
    else {
      $this->getLogger('webform_submission')->error("%error error occurred while appending submission to Google Sheets.", $context);
    }
  }

  /**
   * Helper method to prepare webform submission fields.
   *
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *   A webform submission.
   *
   * @return array
   *   An array with fields.
   */
  private function getData(WebformSubmissionInterface $webform_submission) {
    // Get submission and elements data.
    $data = $webform_submission->toArray(TRUE);

    // Remove unsupported properties from data.
    // These are typically added by other module's like metatag.
    $unsupported_properties = array_combine($this->unsupportedProperties, $this->unsupportedProperties);
    $data = array_diff_key($data, $unsupported_properties);

    // Flatten data and prioritize the element data over the
    // webform submission data.
    $element_data = $data['data'];
    unset($data['data']);
    $data = $element_data + $data;

    // Included selected submission data.
    $data = array_diff_key($data, $this->configuration['included_data']);

    // Append uploaded file name, uri, and base64 data to fields.
    $webform = $this->getWebform();
    foreach ($data as $element_key => $element_value) {
      // Ignore empty and not equal to zero values.
      // @see https://stackoverflow.com/questions/732979/php-whats-an-alternative-to-empty-where-string-0-is-not-treated-as-empty
      if (empty($element_value) && $element_value !== 0 && $element_value !== '0') {
        continue;
      }

      $element = $webform->getElement($element_key);
      if (!$element) {
        continue;
      }

      // Cast markup to string. This only applies to computed Twig values.
      // @see \Drupal\webform\Element\WebformComputedTwig::computeValue
      if ($element_value instanceof MarkupInterface) {
        $data[$element_key] = $element_value = (string) $element_value;
      }

      $element_plugin = $this->elementManager->getElementInstance($element);
      if (!($element_plugin instanceof WebformManagedFileBase)) {
        continue;
      }

      $element_plugin = $this->elementManager->getElementInstance($element);

      if ($element_plugin instanceof WebformManagedFileBase) {
        if ($element_plugin->hasMultipleValues($element)) {
          foreach ($element_value as $fid) {
            $data['_' . $element_key][] = $this->getRequestFileData($fid);
          }
        }
        else {
          $data['_' . $element_key] = $this->getRequestFileData($element_value);
        }
      }
      elseif (!empty($this->configuration['cast'])) {
        // Cast value.
        $data[$element_key] = $this->castRequestValues($element, $element_plugin, $element_value);
      }
    }

    // Replace tokens.
    $data = $this->replaceTokens($data, $webform_submission);

    // Return values only.
    return array_values($data);
  }

  /**
   * Cast request values.
   *
   * @param array $element
   *   An element.
   * @param \Drupal\webform\Plugin\WebformElementInterface $element_plugin
   *   The element's webform plugin.
   * @param mixed $value
   *   The element's value.
   *
   * @return mixed
   *   The element's values cast to boolean or float when appropriate.
   */
  protected function castRequestValues(array $element, WebformElementInterface $element_plugin, $value) {
    $element_plugin->initialize($element);
    if ($element_plugin->hasMultipleValues($element)) {
      foreach ($value as $index => $item) {
        $value[$index] = $this->castRequestValue($element, $element_plugin, $item);
      }
      return $value;
    }
    else {
      return $this->castRequestValue($element, $element_plugin, $value);
    }
  }

  /**
   * Cast request value.
   *
   * @param array $element
   *   An element.
   * @param \Drupal\webform\Plugin\WebformElementInterface $element_plugin
   *   The element's webform plugin.
   * @param mixed $value
   *   The element's value.
   *
   * @return mixed
   *   The element's value cast to boolean or float when appropriate.
   */
  protected function castRequestValue(array $element, WebformElementInterface $element_plugin, $value) {
    if ($element_plugin instanceof BooleanBase) {
      return (boolean) $value;
    }
    elseif ($element_plugin instanceof NumericBase) {
      return (float) $value;
    }
    elseif ($element_plugin instanceof WebformCompositeBase) {
      $composite_elements = (isset($element['#element']))
        ? $element['#element']
        : $element_plugin->getCompositeElements();
      foreach ($composite_elements as $key => $composite_element) {
        if (isset($value[$key])) {
          $composite_element_plugin = $this->elementManager->getElementInstance($composite_element);
          $value[$key] = $this->castRequestValue($composite_element, $composite_element_plugin, $value[$key]);
        }
      }
      return $value;
    }
    else {
      return $value;
    }
  }

  /**
   * Get request file data.
   *
   * @param int $fid
   *   A file id.
   * @param string|null $prefix
   *   A prefix to prepended to data.
   *
   * @return array
   *   An associative array containing file data (name, uri, mime, and data).
   */
  protected function getRequestFileData($fid, $prefix = '') {
    /** @var \Drupal\file\FileInterface $file */
    $file = File::load($fid);
    if (!$file) {
      return [];
    }

    $data = [];
    $data[$prefix . 'id'] = (int) $file->id();
    $data[$prefix . 'name'] = $file->getFilename();
    $data[$prefix . 'uri'] = $file->getFileUri();
    $data[$prefix . 'mime'] = $file->getMimeType();
    $data[$prefix . 'uuid'] = $file->uuid();
    $data[$prefix . 'data'] = base64_encode(file_get_contents($file->getFileUri()));
    return $data;
  }

  /**
   * Helper method to prepare cell object.
   *
   * @param string $data
   *   A string to insert into a cell.
   *
   * @return \Google\Service\Sheets\CellData
   *   Cells to append in the row.
   */
  private function prepareCell($data) {
    $value = new ExtendedValue();

    switch (gettype($data)) {
      case 'boolean':
        $value->setBoolValue($data);
        break;

      case 'integer':
      case 'double':
        $value->setNumberValue($data);
        break;

      case 'string':
        $value->setStringValue($data);
        break;

      default:
        // Set nothing?
        break;
    }

    $cell = new CellData();
    $cell->setUserEnteredValue($value);

    return $cell;
  }

  /**
   * Helper method to prepare row object.
   *
   * @param array $data
   *   An array with row values.
   *
   * @return \Google\Service\Sheets\RowData
   *   Rows to append in the request.
   */
  private function prepareRows(array $data) {
    $row = new RowData();
    $row->setValues(array_map([$this, 'prepareCell'], array_values($data)));

    return $row;
  }

  /**
   * Helper method to prepare Google API request.
   *
   * @param string $sheet_id
   *   A string with Google Spreadsheet Sheet ID.
   * @param \Google\Service\Sheets\RowData $rows
   *   Rows to append in the request.
   *
   * @return \Google\Service\Sheets\BatchUpdateSpreadsheetRequest
   *   Batch update request for a spreadsheet.
   */
  private function prepareRequest($sheet_id, RowData $rows) {
    $append_request = new AppendCellsRequest();
    $append_request->setSheetId($sheet_id);
    $append_request->setRows($rows);
    $append_request->setFields('userEnteredValue');

    $request = new Request();
    $request->setAppendCells($append_request);

    return new BatchUpdateSpreadsheetRequest(['requests' => [$request]]);
  }

  /**
   * Helper method to perform Google API request.
   *
   * @param string $spreadsheet_id
   *   A string with Google Spreadsheet ID.
   * @param \Google\Service\Sheets\BatchUpdateSpreadsheetRequest $request
   *   Batch update request for a spreadsheet.
   *
   * @return \Google\Service\Sheets\BatchUpdateSpreadsheetResponse
   *   Google API response.
   */
  private function makeRequest($spreadsheet_id, BatchUpdateSpreadsheetRequest $request) {
    $google_api_client = $this->getEntityStorage('google_api_client')->load(
      $this->configuration['google_api_client']
    );
    if ($google_api_client) {
      $this->googleService->setGoogleApiClient($google_api_client);
      $service = new Sheets($this->googleService->googleClient);

      return $service->spreadsheets->batchUpdate($spreadsheet_id, $request);
    }
    else {
      throw new Exception(new TranslatableMarkup("Unable to find Google API Entity with ID @id", ['@id' => $this->configuration['google_api_client']]), 1);
    }
  }

}
